from ..identification import IdentificationStrategy

import xarray as xr
from random import randrange
import datetime


class OverlapIdentificationExample(IdentificationStrategy):

    def __init__(self, some_parameter='', **kwargs):
        # Constructor. Called from example_template.py, parameters can be passed and set here.
        self.param = some_parameter

        pass

    def precompute(self, dataset: xr.Dataset, **kwargs):
        # Precomputation, done once for the entire dataset.
        # All work which requires or takes the entire input can be performed here, identify() later is called in
        #   parallel for all timesteps, ensemble members, ...

        # Also data fields need to be added here in case they are needed and populated later on.
        dataset['identified_areas'] = xr.zeros_like(dataset['cv'], dtype=int)
        return dataset

    def identify(self, dataset: xr.Dataset, **kwargs):
        # Identification, here is the place for your identification algorithm.
        # This is called in parallel for each spatial subset of the data, e.g., for each member and for each timestep
        # Therefore, dataset contains only spatial dimensions (lat,lon) if self.processing_mode = '2d' is set,
        #   (see __init__), or (lat,lon,level) if '3d' is set.

        # here your code ... identify objects for this subset.

        # Let's say you detected 5 objects:
        obj_list = []

        cur_time_item = dataset.time.data[0].item()
        dt = datetime.datetime.utcfromtimestamp(cur_time_item / 1e9)

        identified_areas = dataset.identified_areas.isel(plev=0)

        for i in range(5):
            # get an instance of a new object, can pass an ID or set in manually afterwards
            # set some ID to it
            obj = self.get_new_object(id=i+1)

            # identify the areas, simple example
            if dt.hour == 0:
                identified_areas[0:2, 0:2] = obj.id
            elif dt.hour == 6:
                identified_areas[1:3, 1:3] = obj.id
            elif dt.hour == 12:
                identified_areas[4:6, 4:6] = obj.id
            elif dt.hour == 18:
                identified_areas[5:7, 5:7] = obj.id

            # get properties of object and populate them (like defined in template.proto)
            properties = obj.properties

            properties.size = 42
            properties.centroid.x = randrange(10)
            properties.centroid.y = 2 * obj.id
            properties.centroid.z = 3 * obj.id

            properties.list_of_something.append("type x")
            properties.list_of_something.append("type z")

            obj_list.append(obj)

        # return the dataset (can be changed here), and the list of objects
        return dataset, obj_list

    def postprocess(self, dataset: xr.Dataset, obj_desc, **kwargs):
        # Here post process: called once after processing the whole dataset.
        # If needed, the dataset or objects can be changed here.

        # objects are already in finished protobuf description, cf. the JSON output. pb2_desc.sets is list of
        # trackable sets, each one contains the timesteps list, where each entry is a valid time and contains a list
        # of objects.

        # obj_desc = self.make_ids_unique(obj_desc)
        return dataset, obj_desc
